from typing import *
{% for imp in extra_imports %}{{ imp }}
{% endfor %}
import {{to_snake_case(domain.app_name)}}
from {{path_to_module(domain.toolguard_common.file_name)}} import PolicyViolationException
from {{ path_to_module(domain.app_types.file_name) }} import *
from {{ path_to_module(domain.app_api.file_name) }} import {{ domain.app_api_class_name }}

def {{ method.name }}(api: {{ domain.app_api_class_name }}, {{ method.signature }}):
    """
    Policy to check: {{ policy }}

    Args:
        api ({{ domain.app_api_class_name }}): api to access other tools.
{{method.args_doc_str}}
    """
    pass #FIXME
