from typing import Any, Dict, List
from abc import ABC, abstractmethod
from datetime import date, datetime
from {{ types_module }} import *

class {{ class_name }}(ABC):
{% for method in methods %}
    @abstractmethod
    def {{ method.name }}{{ method.signature }}:
        {% if method.doc %}"""{{ method.doc }}"""{% endif %}
        pass
{% endfor %}
